from speedcord import Client
from logging import Logger

from .shard import DefaultShard


class DefaultGatewayHandler:
    client: Client
    logger: Logger

    def __init__(self, client: Client):
        ...

    async def on_receive(self, data: dict, shard: DefaultShard):
        ...



"""
Created by Epic at 9/1/20
"""
from asyncio import Event, get_event_loop, Lock
from logging import getLogger

from .exceptions import Unauthorized, ConnectionsExceeded, InvalidToken
from .http import HttpClient, Route
from .dispatcher import OpcodeDispatcher, EventDispatcher
from .gateway import DefaultGatewayHandler
from .shard import DefaultShard

__all__ = ("Client",)


class Client:
    def __init__(self, intents, token=None, *, shard_count=None, shard_ids=None):
        """
        The client to interact with the discord API
        :param intents: the intents to use
        :param token: the discord bot token to use
        :param shard_count: how many shards to use
        :param shard_ids: A list of shard ids to spawn. Shard_count must be set for this to work
        """
        # Configurable stuff
        self.intents = int(intents)
        self.token = token
        self.shard_count = shard_count
        self.shard_ids = shard_ids

        # Things used by the lib, usually doesn't need to get changed but can if you want to.
        self.shards = []
        self.loop = get_event_loop()
        self.logger = getLogger("speedcord")
        self.http = None
        self.opcode_dispatcher = OpcodeDispatcher(self.loop)
        self.event_dispatcher = EventDispatcher(self.loop)
        self.gateway_handler = DefaultGatewayHandler(self)
        self.connected = Event()
        self.exit_event = Event(loop=self.loop)
        self.remaining_connections = None
        self.connection_lock = Lock(loop=self.loop)

        # Default event handlers
        self.opcode_dispatcher.register(0, self.handle_dispatch)

        # Check types
        if shard_count is None and shard_ids is not None:
            raise TypeError("You have to set shard_count if you use shard_ids")

    def run(self):
        """
        Starts the client
        """
        try:
            self.loop.run_until_complete(self.start())
        except KeyboardInterrupt:
            self.loop.run_until_complete(self.close())

    async def get_gateway(self):
        """
        Get details about the gateway
        :return: wss url to connect to
        :return: how many shards to use
        :return: how many gateway connections left
        :return: how many ms until the gateway connection limit resets
        """
        route = Route("GET", "/gateway/bot")
        try:
            r = await self.http.request(route)
        except Unauthorized:
            await self.close()
            raise
        data = await r.json()

        shards = data["shards"]
        remaining_connections = data["session_start_limit"]["remaining"]
        connections_reset_after = data["session_start_limit"]["reset_after"]
        gateway_url = data["url"]

        if remaining_connections == 0:
            raise ConnectionsExceeded
        self.remaining_connections = remaining_connections
        self.logger.debug(f"{remaining_connections} gateway connections left!")

        return gateway_url, shards, remaining_connections, connections_reset_after

    async def connect(self):
        """
        Connects to discord and spawns shards. Start has to be called first!
        """
        if self.token is None:
            raise InvalidToken

        try:
            gateway_url, shard_count, _, connections_reset_after = await self.get_gateway()
        except Unauthorized:
            self.exit_event.clear()
            raise InvalidToken

        if self.shard_count is None or self.shard_count < shard_count:
            self.shard_count = shard_count

        shard_ids = self.shard_ids or range(self.shard_count)
        for shard_id in shard_ids:
            self.logger.debug(f"Launching shard {shard_id}")
            shard = DefaultShard(shard_id, self, loop=self.loop)
            self.loop.create_task(shard.connect(gateway_url))
            self.shards.append(shard)
        self.connected.set()
        self.logger.info("All shards connected!")

    async def start(self):
        """
        Sets up the http client and connects to discord and spawns shards.
        """
        if self.token is None:
            raise InvalidToken
        self.http = HttpClient(self.token, loop=self.loop)

        await self.connect()

        await self.exit_event.wait()
        await self.close()

    async def close(self):
        """
        Closes the http client and disconnects all shards
        """
        self.connected.clear()
        self.exit_event.set()
        await self.http.close()
        for shard in self.shards:
            await shard.close()

    def listen(self, event):
        """
        Listen to a event or a opcode.
        :param event: a opcode or event name to listen to
        """

        def get_func(func):
            if type(event) == int:
                self.opcode_dispatcher.register(event, func)
            elif type(event) == str:
                self.event_dispatcher.register(event, func)
            else:
                raise TypeError("Invalid event type!")

        return get_func

    # Handle events
    async def handle_dispatch(self, data, shard):
        """
        Dispatches a event to the event handler
        :param data: the data to dispatch
        :param shard: What shard was the event received on
        """
        self.event_dispatcher.dispatch(data["t"], data["d"], shard)




from typing import Optional, Type, Dict, Any
from types import TracebackType
from asyncio import AbstractEventLoop, Lock, Event
from logging import Logger

from aiohttp import ClientWebSocketResponse, ClientResponse, ClientSession


class Route:
    method: str
    path: str
    channel_id: Optional[int]
    guild_id: Optional[int]

    def __init__(self, method: str, route: str, **parameters: Any):
        ...

    @property
    def bucket(self) -> str:
        ...


class LockManager:
    lock: Lock
    unlock: bool

    def __init__(self, lock: Lock):
        ...

    def __enter__(self) -> 'LockManager':
        ...

    def defer(self):
        ...

    def __exit__(self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]):
        ...


class HttpClient:
    baseuri: str
    token: str
    loop: AbstractEventLoop
    session: ClientSession
    logger: Logger
    ratelimit_locks: Dict[str, Lock]
    global_lock: Event
    default_headers: Dict[str, str]
    retry_attempts: int

    def __init__(self, token: str, *, baseuri: str = None, loop: AbstractEventLoop = None):
        ...

    async def create_ws(self, url: str, *, compression: int) -> ClientWebSocketResponse:
        ...

    async def request(self, route: Route, **kwargs: Any) -> ClientResponse:
        ...

    async def close(self):
        ...



from typing import List, Optional, Union, Tuple, Callable, Any
from asyncio import AbstractEventLoop, Event, Lock
from logging import Logger

from .shard import DefaultShard
from .http import HttpClient
from .dispatcher import EventDispatcher, OpcodeDispatcher
from .gateway import DefaultGatewayHandler


class Client:
    intents: int
    token: str
    shard_count: int
    shard_ids: List[int]

    shards: List[DefaultShard]
    loop: AbstractEventLoop
    logger: Logger
    http: HttpClient
    opcode_dispatcher: OpcodeDispatcher
    event_dispatcher: EventDispatcher
    gateway_handler: DefaultGatewayHandler
    connected: Event
    exit_event: Event
    remaining_connections: Optional[int]
    connection_lock: Lock

    def __init__(self, intents: int, token: Optional[str] = None, *, shard_count: Optional[int] = None, shard_ids: Optional[List[int]] = None):
        ...
    def run(self):
        ...
    async def get_gateway(self) -> Tuple[str, int, int, int]:
        ...

    async def connect(self):
        ...
    async def start(self):
        ...
    async def close(self):
        ...

    def listen(self, event: Union[str, int]) -> Callable[[Callable[[dict, DefaultShard], Any]], Any]:
        ...

    async def handle_dispatch(self, data: dict, shard: DefaultShard):
        ...



from speedcord import Client
from typing import Dict, Any, Optional, List, Union


class BaseContext:
    client: Client
    data: Dict[str, Any]

    def __init__(self, client: Client, data: Dict[str, Any]):
        ...


class MessageContext:
    id: int
    channel_id: int
    guild_id: Optional[int]
    author: Dict[str, Any]
    member: Optional[Dict[str, Any]]
    content: str
    timestamp: str
    edited_timestamp: Optional[str]
    tts: bool
    mention_everyone: bool
    mentions: Dict[str, Any]
    mention_roles: List[int]
    mention_channels: Dict[str, Any]
    attachments: List[Dict[str, Any]]
    embeds: List[Dict[str, Any]]
    reactions: List[Dict[str, Any]]
    nonce: Optional[Union[int, str]]
    pinned: bool
    webhook_id: Optional[int]
    type: int
    activity: Optional[Dict[str, Any]]
    application: Optional[Dict[str, Any]]
    message_reference: Optional[Dict[str, int]]
    flags: Optional[int]

    client: Client
    data: Dict[str, Any]

    async def send(self, *, content: str = ..., nonce: Union[int, str] = ..., tts: bool = ...,
                   embed: Dict[str, Any] = ..., allowed_mentions: Dict[str, Any] = ...):
        ...




from typing import Callable, Dict, Any
from asyncio import AbstractEventLoop
from logging import Logger

from .shard import DefaultShard


class OpcodeDispatcher:
    logger: Logger
    loop: AbstractEventLoop

    event_handlers: Dict[int, Callable[[dict, DefaultShard], Any]]

    def __init__(self, loop: AbstractEventLoop):
        ...

    def dispatch(self, opcode: int, *args: Any, **kwargs: Any):
        ...

    def register(self, opcode: int, func: Callable[[dict, DefaultShard], Any]):
        ...


class EventDispatcher:
    logger: Logger
    loop: AbstractEventLoop

    event_handlers: Dict[str, Callable[[dict, DefaultShard], Any]]

    def __init__(self, loop: AbstractEventLoop):
        ...

    def dispatch(self, event_name: str, *args: Any, **kwargs: Any):
        ...

    def register(self, event_name: str, func: Callable[[dict, DefaultShard], Any]):
        ...



from aiohttp import ClientResponse


class HTTPException(Exception):
    request: ClientResponse
    data: str

    def __init__(self, request: ClientResponse, data: str):
        ...


class Forbidden(HTTPException):
    ...


class NotFound(HTTPException):
    request: ClientResponse

    def __init__(self, request: ClientResponse):
        ...


class Unauthorized(HTTPException):
    request: ClientResponse

    def __init__(self, request: ClientResponse):
        ...


class LoginException(Exception):
    ...


class InvalidToken(LoginException):
    def __init__(self):
        ...


class ShardingNotSupported(LoginException):
    def __init__(self):
        ...


class ConnectionsExceeded(LoginException):
    def __init__(self):
        ...


class GatewayException(Exception):
    ...


class GatewayClosed(GatewayException):
    def __init__(self):
        ...



from typing import Optional
from speedcord import Client
from asyncio import AbstractEventLoop, Lock, Event
from logging import Logger

from aiohttp import ClientWebSocketResponse


class DefaultShard:
    id: int
    client: Client
    loop: AbstractEventLoop
    ws: Optional[ClientWebSocketResponse]
    gateway_url: Optional[str]
    ws_ratelimiting_lock: Lock
    logger: Logger
    connected: Event
    received_heartbeat_ack: bool
    heartbeat_interval: Optional[int]
    heartbeat_count: Optional[int]
    failed_heartbeats: int
    session_id: Optional[str]
    last_event_id: Optional[int]

    def __init__(self, shard_id: int, client: Client, loop: AbstractEventLoop):
        ...

    async def connect(self, gateway_url: str):
        ...

    async def close(self):
        ...

    async def read_loop(self):
        ...

    async def send(self, data: dict):
        ...

    async def identify(self):
        ...

    async def resume(self):
        ...

    async def heartbeat_loop(self):
        ...

    async def handle_hello(self, data: dict, shard: 'DefaultShard'):
        ...

    async def handle_heartbeat_ack(self, data: dict, shard: 'DefaultShard'):
        ...

    async def handle_ready(self, data: dict, shard: 'DefaultShard'):
        ...

    async def handle_invalid_session(self, data: dict, shard: 'DefaultShard'):
        ...
